"""A simple and immutable basic statistics record."""

import statistics
from dataclasses import dataclass
from math import isfinite, sqrt, gcd, inf
from typing import Union, Iterable, Final, cast, Dict, Callable

from moptipy.utils.logger import CSV_SEPARATOR, SCOPE_SEPARATOR
from moptipy.utils.math import DBL_INT_LIMIT_P, try_int, try_int_div, \
    try_int_root
from moptipy.utils.strings import num_to_str, str_to_intfloat, \
    str_to_intfloatnone
from moptipy.utils.types import type_error

#: The limit until which we simplify geometric mean data.
_INT_ROOT_LIMIT: Final[int] = int(sqrt(DBL_INT_LIMIT_P))
_ULP: Final[float] = 1 - (2 ** (-53))

#: The minimum value key.
KEY_MINIMUM: Final[str] = "min"
#: The median value key.
KEY_MEDIAN: Final[str] = "med"
#: The arithmetic mean value key.
KEY_MEAN_ARITH: Final[str] = "mean"
#: The geometric mean value key.
KEY_MEAN_GEOM: Final[str] = "geom"
#: The maximum value key.
KEY_MAXIMUM: Final[str] = "max"
#: The standard deviation value key.
KEY_STDDEV: Final[str] = "sd"

#: The number of CSV columns.
CSV_COLS: Final[int] = 6

#: The empty csv row of statistics
EMPTY_CSV_ROW: Final[str] = CSV_SEPARATOR * (CSV_COLS - 1)

#: the internal getters
_GETTERS: Final[Dict[str, Callable[['Statistics'],
                                   Union[int, float, None]]]] = {
    KEY_MINIMUM: lambda s: s.minimum,
    KEY_MEDIAN: lambda s: s.median,
    KEY_MEAN_ARITH: lambda s: s.mean_arith,
    KEY_MEAN_GEOM: lambda s: s.mean_geom,
    KEY_MAXIMUM: lambda s: s.maximum,
    KEY_STDDEV: lambda s: s.stddev
}


@dataclass(frozen=True, init=False, order=True)
class Statistics:
    """An immutable record with statistics of one quantity."""

    #: The minimum.
    minimum: Union[int, float]
    #: The median.
    median: Union[int, float]
    #: The arithmetic mean value.
    mean_arith: Union[int, float]
    #: The geometric mean value, if defined.
    mean_geom: Union[int, float, None]
    #: The maximum.
    maximum: Union[int, float]
    #: The standard deviation.
    stddev: Union[int, float]

    def __init__(self, n: int,
                 minimum: Union[int, float],
                 median: Union[int, float],
                 mean_arith: Union[int, float],
                 mean_geom: Union[int, float, None],
                 maximum: Union[int, float],
                 stddev: Union[int, float]):
        """
        Initialize the statistics class.

        :param minimum: the minimum
        :param median: the median
        :param mean_arith: the arithmetic mean
        :param mean_geom: the geometric mean, or `None` if it is undefined
        :param maximum: the maximum
        :param stddev: the standard deviation (`0` is also used for undefined)
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be >= 1, but is {n}.")

        # check minimum
        if not isinstance(minimum, (int, float)):
            raise type_error(minimum, "minimum", (int, float))
        if isinstance(minimum, float) and (not isfinite(minimum)):
            raise ValueError(f"minimum must be finite, but is {minimum}.")

        # check median
        if not isinstance(median, (int, float)):
            raise type_error(median, "median", (int, float))
        if isinstance(median, float) and (not isfinite(median)):
            raise ValueError(f"med must be finite, but is {median}.")
        if n == 1:
            if median != minimum:
                raise ValueError(f"median ({median}) must equal "
                                 f"minimum ({minimum}) if n=1.")
        elif median < minimum:
            raise ValueError(
                f"median ({median}) must be >= minimum ({minimum}) if n>1.")

        # check maximum
        if not isinstance(maximum, (int, float)):
            raise type_error(maximum, "maximum", (int, float))
        if isinstance(maximum, float) and (not isfinite(maximum)):
            raise ValueError(f"maximum must be finite, but is {maximum}.")
        if n == 1:
            if maximum != minimum:
                raise ValueError(f"maximum ({maximum}) must equal "
                                 f"minimum ({minimum}) if n=1.")
        elif maximum < median:
            raise ValueError(
                f"maximum ({maximum}) must be >= med ({median}) if n>1.")

        # check arithmetic mean
        if not isinstance(mean_arith, (int, float)):
            raise type_error(mean_arith, "mean_arith", (int, float))
        if isinstance(mean_arith, float) and (not isfinite(mean_arith)):
            raise ValueError(
                f"mean_arith must be finite, but is {mean_arith}.")
        if n == 1:
            if mean_arith != minimum:
                raise ValueError(f"mean_arith ({mean_arith}) must equal "
                                 f"minimum ({minimum}) if n=1.")
        else:
            if mean_arith < minimum:
                raise ValueError(
                    f"mean_arith ({mean_arith}) must be >= "
                    f"minimum ({minimum}) if n>1.")
            if mean_arith > maximum:
                raise ValueError(
                    f"mean_arith ({mean_arith}) must be <= "
                    f"maximum ({maximum}) if n>1.")

        # check geometric mean
        if mean_geom is None:
            if minimum > 0:
                raise ValueError(
                    f"If minimum ({minimum}) > 0, then mean_geom must be "
                    f"defined, but it is {mean_geom}.")
        else:
            if minimum <= 0:
                raise ValueError(
                    f"If minimum ({minimum}) <= 0, then mean_geom is "
                    f"undefined, but it is {mean_geom}.")
            if not isinstance(mean_geom, (int, float)):
                raise type_error(mean_geom, "mean_geom", (int, float))
            if isinstance(mean_geom, float) and (not isfinite(mean_geom)):
                raise ValueError(
                    f"mean_geom must be finite, but is {mean_geom}.")
            if n == 1:
                if mean_geom != minimum:
                    raise ValueError(f"mean_geom ({mean_geom}) must equal "
                                     f"minimum ({minimum}) if n=1.")
            else:
                if mean_geom < minimum:
                    raise ValueError(
                        f"mean_geom ({mean_geom}) must be >= "
                        f"minimum ({minimum}) if n>1.")
                if mean_geom > mean_arith:
                    raise ValueError(
                        f"mean_geom ({mean_geom}) must be <= "
                        f"mean_arith ({mean_arith}) if n>1.")
                if (mean_geom >= mean_arith) and (minimum < maximum):
                    raise ValueError(
                        f"mean_geom ({mean_geom}) must be < "
                        f"mean_arith ({mean_arith}) if n>1 and "
                        f"minimum ({minimum}) < maximum ({maximum}).")

        if not isinstance(stddev, (int, float)):
            raise type_error(stddev, "stddev", (int, float))
        if isinstance(stddev, float) and (not isfinite(stddev)):
            raise ValueError(f"stddev must be finite, but is {stddev}.")

        if n == 1:
            if stddev != 0:
                raise ValueError(
                    f"If n==1, stddev must be 0, but is {stddev}.")
            stddev = 0  # set it to int 0

        if (stddev == 0) ^ (minimum == maximum):
            raise ValueError(
                f"If stddev ({stddev}) is 0, then minimum ({minimum}) must "
                f"equal maximum ({maximum}) and vice versa.")
        if minimum == maximum:
            stddev = 0  # set if to int 0

        # fix types to int where possible without loss of precision

        object.__setattr__(self, "minimum", try_int(minimum))
        object.__setattr__(self, "median", try_int(median))
        object.__setattr__(self, "maximum", try_int(maximum))
        object.__setattr__(self, "mean_arith", try_int(mean_arith))
        object.__setattr__(self, "mean_geom",
                           None if mean_geom is None else try_int(mean_geom))
        object.__setattr__(self, "stddev", try_int(stddev))

    def min_mean(self) -> Union[int, float]:
        """
        Obtain the smallest of the three mean values.

        :return: the smallest of `mean_arith`, `mean_geom`, and `median`
        :rtype: Union[int, float]
        """
        if self.mean_geom is None:
            if self.mean_arith < self.median:
                return self.mean_arith
            return self.median

        if self.mean_geom < self.median:
            return self.mean_geom
        return self.median

    def max_mean(self) -> Union[int, float]:
        """
        Obtain the largest of the three mean values.

        :return: the largest of `mean_arith`, `mean_geom`, and `median`
        :rtype: Union[int, float]
        """
        if self.mean_arith > self.median:
            return self.mean_arith
        return self.median

    @staticmethod
    def create(source: Iterable[Union[int, float]]) -> 'Statistics':
        """
        Create a statistics object from an iterable.

        :param source: the source
        :return: a statistics representing the statistics over `source`
        :rtype: Statistics

        >>> from moptipy.evaluation.statistics import Statistics
        >>> s = Statistics.create([3, 1, 2, 5])
        >>> print(s.minimum)
        1
        >>> print(s.maximum)
        5
        >>> print(s.mean_arith)
        2.75
        >>> print(s.median)
        2.5
        >>> print(f"{s.mean_geom:.4f}")
        2.3403
        >>> print(f"{s.min_mean():.4f}")
        2.3403
        >>> print(f"{s.max_mean()}")
        2.75
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)

        minimum: Union[int, float] = inf
        maximum: Union[int, float] = -inf
        can_int: bool = True  # are all values integers?
        int_sum: int = 0  # the integer sum (for mean, stddev)
        int_prod: int = 1  # the integer product (for geom_mean)
        int_sum_sqr: int = 0  # the sum of squares (for stddev)
        n: int = 0  # the number of items

        for e in source:  # iterate over all data
            n = n + 1
            e = try_int(e)

            if can_int:  # can we do integers
                if not isinstance(e, int):
                    can_int = False
            if can_int:  # so far, we only encountered ints
                zz = cast(int, e)
                int_sum += zz  # so we can sum exactly
                int_prod *= zz  # and can compute exact products
                int_sum_sqr += zz * zz

            if e < minimum:
                minimum = e  # update minimum
            if e > maximum:
                maximum = e  # update maximum

        if n <= 0:
            raise ValueError("source cannot be empty.")

        if minimum >= maximum:
            return Statistics(n=n,
                              minimum=minimum,
                              median=minimum,
                              mean_arith=minimum,
                              mean_geom=None if minimum <= 0 else minimum,
                              maximum=maximum,
                              stddev=0)

        stddev: Union[int, float] = 0
        mean_geom: Union[int, float, None] = None
        mean_arith: Union[int, float]

        if can_int:  # if we get here, we have exact sums and product
            mean_arith = try_int_div(int_sum, n)

            if n > 1:  # standard deviation only defined for n > 1
                int_sum2: int = (int_sum * int_sum)
                i_gcd = gcd(int_sum2, n)
                int_sum2 = int_sum2 // i_gcd
                i_n = n // i_gcd

                var: Union[int, float]  # the container for the variance
                if i_n == 1:
                    var = try_int_div(int_sum_sqr - int_sum2, n - 1)
                else:  # variance is float
                    var = (int_sum_sqr - (int_sum2 / i_n)) / (n - 1)

                stddev = try_int(sqrt(var))

            if minimum > 0:  # geometric mean only defined for all-positive
                if int_prod == 0:
                    mean_geom = 0  # geometric mean is 0 if product is 0
                else:  # if not, geom_mean = prod ** (1/n)
                    mean_geom = try_int_root(int_prod, n, True)
                    if mean_geom is None:
                        mean_geom = statistics.geometric_mean(source)
        else:  # ok, we do not have only integer-like values
            mean_arith = try_int(statistics.mean(source))
            if minimum > 0:
                mean_geom = try_int(statistics.geometric_mean(source))
            if n > 1:
                stddev = try_int(statistics.stdev(source))

        if mean_geom is not None:
            # Deal with errors that may have arisen due to
            # numerical imprecision.
            if mean_geom < minimum:
                if (mean_geom / _ULP) >= (minimum * _ULP):
                    mean_geom = minimum
            if mean_geom > maximum:
                if (maximum / _ULP) >= (mean_geom * _ULP):
                    mean_geom = maximum

        return Statistics(n=n,
                          minimum=minimum,
                          median=statistics.median(source),
                          mean_arith=mean_arith,
                          mean_geom=mean_geom,
                          maximum=maximum,
                          stddev=stddev)

    @staticmethod
    def csv_col_names(prefix: str) -> Iterable[str]:
        """
        Make the column names suitable for a CSV-formatted file.

        :param str prefix: the prefix name of the columns
        :return: the column header strings
        :rtype: Iterable[str]
        """
        return [f"{prefix}{SCOPE_SEPARATOR}{b}"
                for b in [KEY_MINIMUM, KEY_MEDIAN,
                          KEY_MEAN_ARITH, KEY_MEAN_GEOM,
                          KEY_MAXIMUM, KEY_STDDEV]]

    @staticmethod
    def value_to_csv(value: Union[int, float]) -> str:
        """
        Expand a single value to a CSV row.

        :param Union[int, float] value: the value
        :return: the CSV row.
        :rtype: str
        """
        s: str = f"{num_to_str(value)}{CSV_SEPARATOR}"
        return f"{s * 5}0"

    def to_csv(self) -> str:
        """
        Generate a string with the data of this record in CSV format.

        :return: the string
        :rtype: str
        """
        q: Final[str] = "" if (self.mean_geom is None) \
            else num_to_str(self.mean_geom)
        return f"{num_to_str(self.minimum)}{CSV_SEPARATOR}" \
               f"{num_to_str(self.median)}{CSV_SEPARATOR}" \
               f"{num_to_str(self.mean_arith)}" \
               f"{CSV_SEPARATOR}" \
               f"{q}" \
               f"{CSV_SEPARATOR}" \
               f"{num_to_str(self.maximum)}{CSV_SEPARATOR}" \
               f"{num_to_str(self.stddev)}"

    @staticmethod
    def from_csv(n: int, row: Union[str, Iterable[str]]) -> 'Statistics':
        """
        Convert a CSV string (or separate CSV cells) to a Statistics object.

        :param int n: the number of observations
        :param Union[str, Iterable[str]] row: either the single string or
            the iterable of separate strings
        :return: the :class:`Statistics` instance
        :rtype: Statistics
        """
        cells = row.split(CSV_SEPARATOR) \
            if isinstance(row, str) else row

        mini, med, mean, geo, maxi, sd = cells
        return Statistics(n,
                          str_to_intfloat(mini),
                          str_to_intfloat(med),
                          str_to_intfloat(mean),
                          str_to_intfloatnone(geo),
                          str_to_intfloat(maxi),
                          str_to_intfloat(sd))

    @staticmethod
    def getter(dimension: str) -> Callable[['Statistics'],
                                           Union[int, float, None]]:
        """
        Produce a function that obtains the given dimension from Statistics.

        :param dimension: the dimension
        :returns: a callable that returns the value corresponding to the
            dimension
        """
        if not isinstance(dimension, str):
            raise type_error(dimension, "dimension", str)
        if dimension in _GETTERS:
            return _GETTERS[dimension]
        raise ValueError(f"unknown dimension '{dimension}', "
                         f"should be one of {sorted(_GETTERS.keys())}.")



"""
In this package, we provide implementations of metaheuristic algorithms.

The black-box methods are given directly in this package, more specialized
algorithms will be placed in sub-packages corresponding to their requirements
for the search- and solution space.
"""



"""A binary operator copying each bit from either source string."""
from typing import Callable, Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op2


class Op2Uniform(Op2):
    """
    This binary search operation copies each bit from either source.

    For each index `i` in the destination array `dest`, uniform
    crossover copies the value from the first source string `x0`with
    probability 0.5 and otherwise the value from the second source
    string `x1`. All bits that have the same value in `x0` and `x1`
    will retain this value in `dest`, all bits where `x0` and `x1`
    differ will effectively be randomized (be `0` with probability 0.5
    and `1` with probability 0.5).
    """

    def op2(self, random: Generator, dest: np.ndarray, x0: np.ndarray,
            x1: np.ndarray) -> None:
        """
        Store result of uniform crossover of `x0` and `x1` in `dest`.

        :param self: the self pointer
        :param random: the random number generator
        :param dest: the array to receive the result
        :param x1: the first existing point in the search space
        :param x2: the second existing point in the search space
        """
        np.copyto(dest, x0)  # copy first source to destination
        ri: Final[Callable[[int], int]] = random.integers
        for i, v in enumerate(x1):  # iterate over second source
            if ri(2) <= 0:  # probability 0.5 = 50% chance to...
                dest[i] = v  # ...copy value from second source

    def __str__(self) -> str:
        """
        Get the name of this binary operator.

        :return: "uniform"
        """
        return "uniform"



"""The multi-objective algorithm execution API."""

from math import isfinite
from typing import Optional, Union, Final, cast

from moptipy.api._mo_process_no_ss import _MOProcessNoSS
from moptipy.api._mo_process_no_ss_log import _MOProcessNoSSLog
from moptipy.api._mo_process_ss import _MOProcessSS
from moptipy.api._mo_process_ss_log import _MOProcessSSLog
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.execution import Execution
from moptipy.api.mo_archive import MOArchivePruner, check_mo_archive_pruner
from moptipy.api.mo_problem import MOProblem, check_mo_problem, \
    MOSOProblemBridge
from moptipy.api.mo_process import MOProcess
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import check_max_fes, check_max_time_millis, \
    check_goal_f
from moptipy.api.space import Space
from moptipy.api.space import check_space
from moptipy.mo.archive.keep_farthest import KeepFarthest
from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.types import type_error


class MOExecution(Execution):
    """
    Define all the components of a multi-objective experiment and execute it.

    Different from :class:`~moptipy.api.execution.Execution`, this class here
    allows us to construct multi-objective optimization processes, i.e., such
    that have more than one optimization goal.
    """

    def __init__(self) -> None:
        """Create the multi-objective execution."""
        super().__init__()
        #: the maximum size of a pruned archive
        self._archive_max_size: Optional[int] = None
        #: the archive size limit at which pruning should be performed
        self._archive_prune_limit: Optional[int] = None
        #: the archive pruning strategy
        self._archive_pruner: Optional[MOArchivePruner] = None

    def set_archive_max_size(self, size: int) -> 'MOExecution':
        """
        Set the upper limit for the archive size (after pruning).

        The internal archive of the multi-objective optimization process
        retains non-dominated solutions encountered during the search. Since
        there can be infinitely many such solutions, the archive could grow
        without bound if left untouched.
        Therefore, we define two size limits: the maximum archive size
        (defined by this method) and the pruning limit. Once the archive grows
        beyond the pruning limit, it is cut down to the archive size limit.

        :param size: the maximum archive size
        :returns: this execution
        """
        if not isinstance(size, int):
            raise type_error(size, "size", int)
        if size <= 0:
            raise ValueError(
                f"archive max size must be positive, but is {size}.")
        if self._archive_prune_limit is not None:
            if size > self._archive_prune_limit:
                raise ValueError(
                    f"archive max size {size} must be <= than archive "
                    f"prune limit {self._archive_prune_limit}")
        self._archive_max_size = size
        return self

    def set_archive_pruning_limit(self, limit: int) -> 'MOExecution':
        """
        Set the size limit of the archive above which pruning is performed.

        If the size of the archive grows above this limit, the archive will be
        pruned down to the archive size limit.

        :param limit: the archive pruning limit
        :returns: this execution
        """
        if not isinstance(limit, int):
            raise type_error(limit, "limit", int)
        if limit <= 0:
            raise ValueError(
                f"archive pruning limit must be positive, but is {limit}.")
        if self._archive_max_size is not None:
            if limit < self._archive_max_size:
                raise ValueError(
                    f"archive pruning limit {limit} must be >= than archive "
                    f"maximum size {self._archive_max_size}")
        self._archive_prune_limit = limit
        return self

    def set_archive_pruner(self, pruner: MOArchivePruner) -> 'MOExecution':
        """
        Set the pruning strategy for downsizing the archive.

        :param pruner: the archive pruner
        :returns: this execution
        """
        self._archive_pruner = check_mo_archive_pruner(pruner)
        return self

    def set_objective(self, objective: Objective) -> 'MOExecution':
        """
        Set the objective function in form of a multi-objective problem.

        :param objective: the objective function
        :returns: this execution
        """
        check_objective(objective)
        if not isinstance(objective, MOProblem):
            objective = MOSOProblemBridge(objective)
        super().set_objective(check_mo_problem(objective))
        return self

    def set_algorithm(self, algorithm: Algorithm) -> 'MOExecution':
        """
        Set the algorithm to be used for this experiment.

        :param algorithm: the algorithm
        :returns: this execution
        """
        super().set_algorithm(algorithm)
        return self

    def set_solution_space(self, solution_space: Space) -> 'MOExecution':
        """
        Set the solution space to be used for this experiment.

        This is the space managing the data structure holding the candidate
        solutions.

        :param solution_space: the solution space
        :returns: this execution
        """
        super().set_solution_space(solution_space)
        return self

    def set_search_space(self, search_space: Optional[Space]) -> 'MOExecution':
        """
        Set the search space to be used for this experiment.

        This is the space from which the algorithm samples points.

        :param search_space: the search space, or `None` of none shall be
            used, i.e., if search and solution space are the same
        :returns: this execution
        """
        super().set_search_space(search_space)
        return self

    def set_encoding(self, encoding: Optional[Encoding]) -> 'MOExecution':
        """
        Set the encoding to be used for this experiment.

        This is the function translating from the search space to the
        solution space.

        :param encoding: the encoding, or `None` of none shall be used
        :returns: this execution
        """
        super().set_encoding(encoding)
        return self

    def set_rand_seed(self, rand_seed: Optional[int]) -> 'MOExecution':
        """
        Set the seed to be used for initializing the random number generator.

        :param rand_seed: the random seed, or `None` if a seed should
            automatically be chosen when the experiment is executed
        """
        super().set_rand_seed(rand_seed)
        return self

    def set_max_fes(self, max_fes: int,
                    force_override: bool = False) -> 'MOExecution':
        """
        Set the maximum FEs.

        This is the number of candidate solutions an optimization is allowed
        to evaluate. If this method is called multiple times, then the
        shortest limit is used unless `force_override` is `True`.

        :param max_fes: the maximum FEs
        :param force_override: the use the value given in `max_time_millis`
            regardless of what was specified before
        :returns: this execution
        """
        super().set_max_fes(max_fes, force_override)
        return self

    def set_max_time_millis(self, max_time_millis: int,
                            force_override: bool = False) -> 'MOExecution':
        """
        Set the maximum time in milliseconds.

        This is the maximum time that the process is allowed to run. If this
        method is called multiple times, the shortest time is used unless
        `force_override` is `True`.

        :param max_time_millis: the maximum time in milliseconds
        :param force_override: the use the value given in `max_time_millis`
            regardless of what was specified before
        :returns: this execution
        """
        super().set_max_time_millis(max_time_millis, force_override)
        return self

    def set_goal_f(self, goal_f: Union[int, float]) -> 'MOExecution':
        """
        Set the goal objective value after which the process can stop.

        If this method is called multiple times, then the largest value is
        retained.

        :param goal_f: the goal objective value.
        :returns: this execution
        """
        super().set_goal_f(goal_f)
        return self

    def set_log_file(self, log_file: Optional[str]) -> 'MOExecution':
        """
        Set the log file to write to.

        This method can be called arbitrarily often.

        :param log_file: the log file
        """
        super().set_log_file(log_file)
        return self

    def set_log_improvements(self, log_improvements: bool = True) \
            -> 'MOExecution':
        """
        Set whether improvements should be logged.

        :param log_improvements: if improvements should be logged?
        :returns: this execution
        """
        super().set_log_improvements(log_improvements)
        return self

    def set_log_all_fes(self, log_all_fes: bool = True) -> 'MOExecution':
        """
        Set whether all FEs should be logged.

        :param log_all_fes: if all FEs should be logged?
        :returns: this execution
        """
        super().set_log_all_fes(log_all_fes)
        return self

    def execute(self) -> MOProcess:
        """
        Create a multi-objective process, apply algorithm, and return result.

        This method is multi-objective equivalent of the
        :meth:`~moptipy.api.execution.Execution.execute` method. It returns a
        multi-objective process after applying the multi-objective algorithm.

        :returns: the instance of :class:`~moptipy.api.mo_process.MOProcess`
            after applying the algorithm.
        """
        objective: Final[MOProblem] = cast(MOProblem, self._objective)
        solution_space: Final[Space] = check_space(self._solution_space)
        search_space: Final[Optional[Space]] = check_space(
            self._search_space, self._encoding is None)
        encoding: Final[Optional[Encoding]] = check_encoding(
            self._encoding, search_space is None)
        rand_seed = self._rand_seed
        if not (rand_seed is None):
            rand_seed = rand_seed_check(rand_seed)
        max_time_millis = check_max_time_millis(self._max_time_millis, True)
        max_fes = check_max_fes(self._max_fes, True)
        goal_f = check_goal_f(self._goal_f, True)
        f_lb = objective.lower_bound()
        if (not (f_lb is None)) and isfinite(f_lb):
            if (goal_f is None) or (f_lb > goal_f):
                goal_f = f_lb

        log_all_fes = self._log_all_fes
        log_improvements = self._log_improvements or self._log_all_fes

        log_file = self._log_file
        if log_file is None:
            if log_all_fes:
                raise ValueError("Log file cannot be None "
                                 "if all FEs should be logged.")
            if log_improvements:
                raise ValueError("Log file cannot be None "
                                 "if improvements should be logged.")
        else:
            log_file.create_file_or_truncate()

        pruner: Final[MOArchivePruner] = \
            self._archive_pruner if self._archive_pruner is not None \
            else KeepFarthest(objective)
        dim: Final[int] = objective.f_dimension()
        size: Final[int] = self._archive_max_size if \
            self._archive_max_size is not None else (
            self._archive_prune_limit if
            self._archive_prune_limit is not None
            else (1 if dim == 1 else 32))
        limit: Final[int] = self._archive_prune_limit if \
            self._archive_prune_limit is not None \
            else (1 if dim == 1 else (size * 4))
        algorithm: Final[Algorithm] = check_algorithm(self._algorithm)

        process: _MOProcessNoSS
        if search_space is None:
            if log_improvements or log_all_fes:
                process = _MOProcessNoSSLog(
                    solution_space, objective, algorithm, pruner, size, limit,
                    log_file, rand_seed, max_fes, max_time_millis, goal_f,
                    log_all_fes)
            else:
                process = _MOProcessNoSS(
                    solution_space, objective, algorithm, pruner, size, limit,
                    log_file, rand_seed, max_fes, max_time_millis, goal_f)
        else:
            if log_improvements or log_all_fes:
                process = _MOProcessSSLog(
                    solution_space, objective, algorithm, pruner, size, limit,
                    log_file, search_space, encoding, rand_seed, max_fes,
                    max_time_millis, goal_f, log_all_fes)
            else:
                process = _MOProcessSS(
                    solution_space, objective, algorithm, pruner, size, limit,
                    log_file, search_space, encoding, rand_seed, max_fes,
                    max_time_millis, goal_f)

        try:
            # noinspection PyProtectedMember
            process._after_init()
            algorithm.solve(process)
        except BaseException as be:
            # noinspection PyProtectedMember
            process._caught = be
        return process



"""A set of utilities interactions with the file system."""
import os
from contextlib import AbstractContextManager
from shutil import rmtree
from tempfile import mkstemp, mkdtemp
from typing import Optional

from moptipy.utils.path import Path


class TempDir(Path, AbstractContextManager):
    """
    A scoped temporary directory to be used in a 'with' block.

    The directory and everything in it will be deleted upon exiting the
    'with' block.
    """

    #: is the directory open?
    __is_open: bool

    def __new__(cls, value):
        """
        Construct the object.

        :param value: the string value
        """
        ret = super(TempDir, cls).__new__(cls, value)
        ret.enforce_dir()
        ret.__is_open = True
        return ret

    @staticmethod
    def create(directory: Optional[str] = None) -> 'TempDir':
        """
        Create the temporary directory.

        :param directory: an optional root directory
        :raises TypeError: if `directory` is not `None` but also no `str`
        """
        if not (directory is None):
            root_dir = Path.path(directory)
            root_dir.enforce_dir()
        else:
            root_dir = None
        return TempDir(mkdtemp(dir=root_dir))

    def __enter__(self) -> 'TempDir':
        """Nothing, just exists for `with`."""
        if not self.__is_open:
            raise ValueError(f"Temporary directory '{self}' already closed.")
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Delete the temporary directory and everything in it.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        opn = self.__is_open
        self.__is_open = False
        if opn:
            rmtree(self, ignore_errors=True, onerror=None)
        return exception_type is None


class TempFile(Path, AbstractContextManager):
    """
    A scoped temporary file to be used in a 'with' block.

    This file will be deleted upon exiting the 'with' block.
    """

    #: is the directory open?
    __is_open: bool

    def __new__(cls, value):
        """
        Construct the object.

        :param value: the string value
        """
        ret = super(TempFile, cls).__new__(cls, value)
        ret.enforce_file()
        ret.__is_open = True
        return ret

    @staticmethod
    def create(directory: Optional[str] = None,
               prefix: Optional[str] = None,
               suffix: Optional[str] = None) -> 'TempFile':
        """
        Create a temporary file.

        :param directory: a root directory or `TempDir` instance
        :param prefix: an optional prefix
        :param suffix: an optional suffix, e.g., `.txt`
        :raises TypeError: if any of the parameters does not fulfill the type
            contract
        """
        if prefix is not None:
            prefix = prefix.strip()
            if not prefix:
                raise ValueError("Prefix cannot be empty if specified.")

        if suffix is not None:
            suffix = suffix.strip()
            if not suffix:
                raise ValueError("Prefix cannot be empty if specified.")

        if directory is not None:
            base_dir = Path.path(directory)
            base_dir.enforce_dir()
        else:
            base_dir = None

        (handle, path) = mkstemp(suffix=suffix, prefix=prefix, dir=base_dir)
        os.close(handle)
        return TempFile(path)

    def __enter__(self) -> 'TempFile':
        """Nothing, just exists for `with`."""
        if not self.__is_open:
            raise ValueError(f"Temporary file '{self}' already deleted.")
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Delete the temporary file.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        opn = self.__is_open
        self.__is_open = False
        if opn:
            os.remove(self)
        return exception_type is None



"""Some internal utilities for optimization."""
from typing import Final, Union


# start record
class Record:
    """
    A point in the search space, its quality, and creation time.

    A record stores a point in the search space :attr:`x` together with
    the corresponding objective value :attr:`f`. It also stores a
    "iteration index" :attr:`it`, i.e., the time when the point was
    created or modified.

    This allows for representing and storing solutions in a population.
    If the population is sorted, then individual records with better
    objective value will be moved to the beginning of the list. Ties are
    broken such that younger individuals (with higher :attr:`it` value)
    are preferred.
    """

    def __init__(self, x, f: Union[int, float]):
        """
        Create the record.

        :param x: the data structure for a point in the search space
        :param f: the corresponding objective value
        """
        #: the point in the search space
        self.x: Final = x
        #: the objective value corresponding to x
        self.f: Union[int, float] = f
        #: the iteration index when the record was created
        self.it: int = 0

    def __lt__(self, other):
        """Precedence if 1) better or b) equally good but younger."""
        f1: Final[Union[int, float]] = self.f
        f2: Final[Union[int, float]] = other.f
        return (f1 < f2) or ((f1 == f2) and (self.it > other.it))
# end record

    def __str__(self) -> str:
        """
        Return the string representation of this object.

        :returns: the string representation of this object.
        """
        return f"({self.f}, {self.it})"


def _int_0(_: int) -> int:
    """
    Do nothing.

    :retval 0: always
    """
    return 0


def _float_0() -> float:
    """
    Do nothing.

    :retval 0.0: always
    """
    return 0.0



"""
The implementation of a (mu+lambda) Evolutionary Algorithm.

This is the basic `mu+lambda`-EA works as follows:

1. Start with a list `lst` of `mu` random records and `lambda` blank records.
2. In each iteration:
    2.1. Use the `mu` first records as input to the search operators to
         generate `lambda` new points in the search space:
         For each new point to be created, the binary operator is applied
         with probability `0<=br<=1` and the unary operator is used otherwise.

         Each of the `mu` records has the same chance to produce such a new
         point in the search space, but no record can be used again (in one
         iteration) until all other `mu-1` selected records have produced at
         least one point. In other words, if `lambda > mu`, then each of the
         `mu` selected record will produce at least `lambda // mu` offspring
         and at most `1 + lambda // mu`.
    2.1. Sort the list `lst` according to the objective value of the record.
         Ties are broken such that younger record are preferred over older
         ones (if they have the same objective value).

If `mu=1`, `lambda=1`, and `br=0`, then this algorithm is exactly equivalent
to the :class:`~moptipy.algorithms.rls.RLS` if the same unary and nullary
operator are used. It is only a bit slower due to the additional overhead of
maintaining a list of records.
"""
from math import isfinite
from typing import Final, Union, Callable, List, cast, Optional

from numpy.random import Generator

from moptipy.algorithms.utils import Record, _int_0, _float_0
from moptipy.api.algorithm import Algorithm2
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import float_to_str
from moptipy.utils.types import type_error


# start nobinary
class EA(Algorithm2):
    """
    The EA is a population-based algorithm using unary and binary operators.

    It starts with a list of `mu` randomly initialized solutions. In each
    step, it retains the `mu` best solutions and generates `lambda` new
    solutions from them using the unary or binary search operator. From the
    joint set of `mu+lambda` solutions, it again selects the best `mu` ones
    for the next iteration. And so on.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the EA to an optimization problem.

        :param process: the black-box process object
        """
        mu: Final[int] = self.__mu  # mu: number of best solutions kept
        lst_size: Final[int] = mu + self.__lambda  # size = mu + lambda
        # Omitted for brevity: store function references in variables
        # end nobinary
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        br: Final[float] = self.__br  # the application are of binary op
        should_terminate: Final[Callable] = process.should_terminate
        r0i: Final[Callable[[int], int]] = cast(   # only if m > 1, we
            Callable[[int], int], random.integers  # need random
            if mu > 1 else _int_0)                 # indices
        r01: Final[Callable[[], float]] = cast(  # only if 0<br<1, we
            Callable[[], float],                 # need random floats
            random.random if 0 < br < 1 else _float_0)
        # start nobinary
        # create list of mu random records and lambda empty records
        lst: Final[List] = [None] * lst_size  # pre-allocate list
        f: Union[int, float] = 0  # variable to hold objective values
        for i in range(lst_size):  # fill list of size mu+lambda
            x = create()  # by creating point in search space
            if i < mu:  # only the first mu records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?
                    return   # computational budget exhausted -> quit
                f = evaluate(x)  # continue? ok, evaluate new solution
            lst[i] = Record(x, f)  # create and store record

        iteration: int = 1  # The first "real" iteration has index 1
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            end: int = mu  # the end index for available old records
            for oi in range(mu, lst_size):  # for all lambda offspring
                if should_terminate():  # only continue if we still...
                    return  # have sufficient budget ... otherwise quit

                offspring: Record = lst[oi]  # pick offspring
                x = offspring.x  # the point in search space we work on
                offspring.it = iteration  # mark as member of new iter.

                pi: int = r0i(end)  # random record in 0..end-1=unused
                rec: Record = lst[pi]  # pick the unused record
                if end > 1:  # swap rec to the end to not use again
                    end = end - 1  # reduce number of unused records
                    lst[end], lst[pi] = rec, lst[end]  # rec <-> end
                else:  # oh: we have used all records
                    end = mu  # then lambda >= mu and we re-use them
                # end nobinary
                # start binary
                if r01() < br:  # apply binary operator at rate br
                    pi = r0i(end)  # random record in 0..end-1
                    rec2: Record = lst[pi]  # pick second unused record
                    if end > 1:  # swap rec2 to end to not use again
                        end = end - 1  # reduce num. of unused records
                        lst[end], lst[pi] = rec2, lst[end]
                    else:  # oh: we have used all records?
                        end = mu  # then lambda>=mu and we re-use them
                    op2(random, x, rec.x, rec2.x)  # apply binary op
                    offspring.f = evaluate(x)  # evaluate new point
                    continue  # below is "else" part with unary operat.
                # end binary
                # start nobinary
                op1(random, x, rec.x)  # apply unary operator to rec.x
                offspring.f = evaluate(x)  # evaluate new point

            iteration = iteration + 1  # step iteration counter
            lst.sort()  # sort list: best records come first
# end nobinary

    def __init__(self, op0: Op0,
                 op1: Optional[Op1] = None,
                 op2: Optional[Op2] = None,
                 mu: int = 1, lambda_: int = 1,
                 br: Optional[float] = None) -> None:
        """
        Create the Evolutionary Algorithm (EA) without binary crossover.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        :param br: the rate at which the binary operator is applied
        """
        if op1 is None:
            op1 = Op1()
            if br is None:
                br = 1.0
            elif br != 1.0:
                raise ValueError(
                    f"if op1==None, br must be None or 1.0, but is {br}.")
        elif op1.__class__ == Op1:
            if br is None:
                br = 1.0
            elif br != 1.0:
                raise ValueError(
                    f"if op1 is Op1, br must be None or 1.0, but is {br}.")
        elif (br is not None) and (br == 1.0):
            op1 = Op1()

        if op2 is None:
            op2 = Op2()
            if br is None:
                br = 0.0
            elif br != 0.0:
                raise ValueError(
                    f"if op2==None, br must be None or 0.0, but is {br}.")
        elif op2.__class__ == Op2:
            if br is None:
                br = 0.0
            elif br != 0.0:
                raise ValueError(
                    f"if op2 is Op2, br must be None or 0.0, but is {br}.")
        elif (br is not None) and (br == 0.0):
            op2 = Op2()
        elif mu == 1:
            if br is None:
                br = 0.0
                op2 = Op2()

        if br is None:
            br = 0.2

        super().__init__(
            f"ea_{mu}_{lambda_}_{float_to_str(br).replace('.', 'd')}"
            if 0 < br < 1 else f"ea_{mu}_{lambda_}", op0, op1, op2)

        if not isinstance(mu, int):
            raise type_error(mu, "mu", int)
        if not (0 < mu <= 1_000_000):
            raise ValueError(f"invalid mu={mu}, must be in 1..1_000_000.")
        if not isinstance(lambda_, int):
            raise type_error(lambda_, "lambda", int)
        if not (0 < lambda_ <= 1_000_000):
            raise ValueError(
                f"invalid lambda={lambda_}, must be in 1..1_000_000.")
        if not isinstance(br, float):
            raise type_error(br, "br", float)
        if not (isfinite(br) and (0.0 <= br <= 1.0)):
            raise ValueError(f"invalid br={br}, must be in [0, 1].")
        if (br > 0.0) and (mu <= 1):
            raise ValueError(
                f"if br (={br}) > 0, then mu (={mu}) must be > 1.")
        #: the number of records to survive in each generation
        self.__mu: Final[int] = mu
        #: the number of offsprings per generation
        self.__lambda: Final[int] = lambda_
        #: the rate at which the binary operator is applied
        self.__br: Final[float] = br

    def log_parameters_to(self, logger: KeyValueLogSection):
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("mu", self.__mu)
        logger.key_value("lambda", self.__lambda)
        logger.key_value("br", self.__br)



"""Generate random mock experiment parameters."""

from dataclasses import dataclass
from math import isfinite, ceil
from typing import Tuple, Final, Set, List, Optional, Iterable, Union, \
    Any, Dict

from numpy.random import Generator

from moptipy.utils.console import logger
from moptipy.utils.nputils import rand_generator, rand_seeds_from_str
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_error


def fixed_random_generator() -> Generator:
    """
    Get the single, fixed random generator for the dummy experiment API.

    :returns: the random number generator
    """
    if not hasattr(fixed_random_generator, 'gen'):
        setattr(fixed_random_generator, "gen", rand_generator(1))
    return getattr(fixed_random_generator, 'gen')


def _random_name(namelen: int,
                 random: Generator = fixed_random_generator()) -> str:
    """
    Generate a random name of a given length.

    :param namelen: the length of the name
    :param random: a random number generator
    :returns: a name of the length
    """
    if not isinstance(namelen, int):
        raise type_error(namelen, "namelen", int)
    if namelen <= 0:
        raise ValueError(f"namelen must be > 0, but is {namelen}.")
    namer: Final[Tuple[str, str, str]] = ("bcdfghjklmnpqrstvwxyz", "aeiou",
                                          "0123456789")
    name = ["x"] * namelen
    index: int = 0
    n_done: bool = False
    for i in range(namelen):
        namee = namer[index]
        name[i] = namee[random.integers(len(namee))]
        n_done = n_done or (i == 2)
        if index <= 0:
            index += 1
        else:
            if index == 2:
                if random.integers(2) <= 0:
                    continue
            index = int((index + 1 + random.integers(2)) % len(namer))
            if n_done:
                while index == 2:
                    index = 1 - min(1, random.integers(6))

    return "".join(name)


def __append_not_allowed(forbidden,
                         dest: Set[Union[str, float, int]]):
    """
    Append items to the set of not-allowed values.

    :param forbidden: the forbidden elements
    :param dest: the set to append to
    """
    if not forbidden:
        return
    if isinstance(forbidden, (str, int, float)):
        dest.add(forbidden)
    elif isinstance(forbidden, Instance):
        dest.add(forbidden.name)
        dest.add(forbidden.hardness)
        dest.add(forbidden.jitter)
        dest.add(forbidden.scale)
        dest.add(forbidden.best)
        dest.add(forbidden.worst)
    elif isinstance(forbidden, Algorithm):
        dest.add(forbidden.name)
        dest.add(forbidden.strength)
        dest.add(forbidden.jitter)
    elif isinstance(forbidden, Iterable):
        for item in forbidden:
            __append_not_allowed(item, dest)
    else:
        raise type_error(forbidden, "element to add",
                         (str, int, float, Instance, Algorithm, Iterable))


def _make_not_allowed(forbidden: Optional[Iterable[Union[
    str, float, int, Iterable[Any]]]] = None) -> \
        Set[Union[str, float, int]]:
    """
    Create a set of not-allowed values.

    :param forbidden: the forbidden elements
    :returns: the set of not-allowed values
    """
    not_allowed: Set[Any] = set()
    __append_not_allowed(forbidden, not_allowed)
    return not_allowed


@dataclass(frozen=True, init=False, order=True)
class Instance:
    """An immutable instance description record."""

    #: The instance name.
    name: str
    #: The instance hardness, in (0, 1), larger values are worst
    hardness: float
    #: The instance jitter, in (0, 1), larger values are worst
    jitter: float
    #: The instance scale, in (0, 1), larger values are worst
    scale: float
    #: The best (smallest) possible objective value
    best: int
    #: The worst (largest) possible objective value
    worst: int
    #: The set of attractors, i.e., local optima - including best and worst
    attractors: Tuple[int, ...]

    def __init__(self,
                 name: str,
                 hardness: float,
                 jitter: float,
                 scale: float,
                 best: int,
                 worst: int,
                 attractors: Tuple[int, ...]):
        """
        Create a mock problem instance description.

        :param name: the instance name
        :param hardness: the instance hardness
        :param jitter: the instance jitter
        :param scale: the instance scale
        :param best: the best (smallest) possible objective value
        :param worst: the worst (largest) possible objective value
        :param attractors: the set of attractors, i.e., local
            optima - including best and worst
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        if name != sanitize_name(name):
            raise ValueError(f"Invalid name '{name}'.")
        object.__setattr__(self, "name", name)

        if not isinstance(hardness, float):
            raise type_error(hardness, "hardness", float,)
        if (not isfinite(hardness)) or (hardness <= 0) or (hardness >= 1):
            raise ValueError(
                f"hardness must be in (0, 1), but is {hardness}.")
        object.__setattr__(self, "hardness", hardness)

        if not isinstance(jitter, float):
            raise type_error(jitter, "jitter", float)
        if (not isfinite(jitter)) or (jitter <= 0) or (jitter >= 1):
            raise ValueError(
                f"jitter must be in (0, 1), but is {jitter}.")
        object.__setattr__(self, "jitter", jitter)

        if not isinstance(scale, float):
            raise type_error(scale, "scale", float)
        if (not isfinite(scale)) or (scale <= 0) or (scale >= 1):
            raise ValueError(
                f"scale must be in (0, 1), but is {scale}.")
        object.__setattr__(self, "scale", scale)

        if not isinstance(best, int):
            raise type_error(best, "best", int)
        if (best <= 0) or (best >= 1_000_000_000):
            raise ValueError(
                f"best must be in 1...999999999, but is {best}.")
        object.__setattr__(self, "best", best)

        if not isinstance(worst, int):
            raise type_error(worst, "worst", int)
        if (worst <= (best + 7)) or (worst >= 1_000_000_000):
            raise ValueError(
                f"worst must be in {best + 8}...999999999, but is {worst}.")
        object.__setattr__(self, "worst", worst)

        if not isinstance(attractors, tuple):
            raise type_error(attractors, "attractors", Tuple)
        if len(attractors) < 4:
            raise ValueError("attractors must contain at least 2 values,"
                             f" but contains only {len(attractors)}.")
        if attractors[0] != best:
            raise ValueError(
                f"attractors[0] must be {best}, but is {attractors[0]}")
        if attractors[-1] != worst:
            raise ValueError(
                f"attractors[-1] must be {worst}, but is {attractors[-1]}")
        prev = -1
        for att in attractors:
            if not isinstance(att, int):
                raise type_error(att, "each attractor", int)
            if (att < best) or (att > worst) or (att <= prev):
                raise ValueError(f"{att} not permitted after {prev} "
                                 f"for best={best} and worst={worst}.")
            prev = att
        object.__setattr__(self, "attractors", attractors)

    @staticmethod
    def create(n: int,
               forbidden: Optional[Any] = None,
               random: Generator = fixed_random_generator()) \
            -> Tuple['Instance', ...]:
        """
        Create a set of fixed problem instances.

        :param n: the number of instances to generate
        :param random: a random number generator
        :param forbidden: the forbidden names and hardnesses
        :returns: a tuple of instances
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be > 0, but is {n}.")
        logger(f"now creating {n} instances.")

        not_allowed: Final[Set[Union[str, float]]] = \
            _make_not_allowed(forbidden)
        names: List[str] = []
        hardnesses: List[float] = []
        jitters: List[float] = []
        scales: List[float] = []

        # First we choose a unique name.
        max_name_len: int = int(max(2, ceil(n / 6)))
        trials: int = 0
        while len(names) < n:
            trials += 1
            if trials > 1000:
                trials = 1
                max_name_len += 1
            nv = _random_name(int(3 + random.integers(max_name_len)), random)
            if nv not in not_allowed:
                names.append(nv)
                not_allowed.add(nv)

        # Now we pick an instance hardness.
        limit: int = 2000
        trials = 0
        while len(hardnesses) < n:
            v: float = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5, scale=0.25) * limit) / limit
            if v not in not_allowed:
                hardnesses.append(v)
                not_allowed.add(v)

        # Now we pick an instance jitter.
        limit = 2000
        trials = 0
        while len(jitters) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5, scale=0.2) * limit) / limit
            if v not in not_allowed:
                jitters.append(v)
                not_allowed.add(v)

        # Now we choose a scale.
        limit = 2000
        trials = 0
        while len(scales) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5, scale=0.2) * limit) / limit
            if v not in not_allowed:
                scales.append(v)
                not_allowed.add(v)

        # We choose the global optimum and the worst objective value.
        trials = 0
        scale: int = 1000
        loc: int = 1000
        limits: List[int] = []
        while len(limits) < (2 * n):
            tbound: int = -1
            while (tbound <= 0) or (tbound >= 1_000_000_000):
                trials += 1
                if trials > 1000:
                    trials = 0
                    scale += (scale // 3)
                    loc *= 2
                tbound = int(random.normal(
                    loc=random.integers(low=1, high=4) * loc, scale=scale))
            permitted: bool = True
            for b in limits:
                if abs(b - tbound) <= 41:
                    permitted = False
                    break
            if permitted and (tbound not in not_allowed):
                limits.append(tbound)
                not_allowed.add(tbound)

        result: List[Instance] = []
        attdone: Set[int] = set()

        for i in range(n, 0, -1):
            trials = 0
            b1 = b2 = -1
            while trials < 10:
                trials += 1
                b1 = limits.pop(random.integers(i * 2))
                b2 = limits.pop(random.integers(i * 2 - 1))
                if b1 > b2:
                    b1, b2 = b2, b1
                if i <= 1:
                    break
                if (b1 * max(2.0, 0.4 * (10 - trials))) < b2:
                    break
                limits.append(b1)
                limits.append(b2)
            attdone.clear()
            attdone.add(b1)
            attdone.add(b2)

            # Now we make sure that there are at least 6 attractors.
            trials = 0
            min_dist = max(7.0, 0.07 * (b2 - b1))
            while ((len(attdone) < 6) or (random.integers(7) > 0)) \
                    and (trials < 20000):
                a = -1
                while ((a <= b1) or (a >= b2) or (a in attdone)) \
                        and (trials < 20000):
                    trials += 1
                    a = int(random.integers(low=b1 + 1, high=b2 - 1))
                if (a <= b1) or (a >= b2):
                    continue
                ok = True
                for aa in attdone:
                    if abs(aa - a) < min_dist:
                        ok = False
                        break
                if ok:
                    attdone.add(a)
                else:
                    if (trials % 1000) <= 0:
                        min_dist = max(5, 0.5 * min_dist)

            result.append(Instance(
                name=names.pop(random.integers(i)),
                hardness=hardnesses.pop(random.integers(i)),
                jitter=jitters.pop(random.integers(i)),
                scale=scales.pop(random.integers(i)),
                best=b1,
                worst=b2,
                attractors=tuple(sorted(attdone))))
        result.sort()
        logger(f"finished creating {n} instances.")
        return tuple(result)


@dataclass(frozen=True, init=False, order=True)
class Algorithm:
    """An immutable algorithm description record."""

    #: The algorithm name.
    name: str
    #: The algorithm strength, in (0, 1), larger values are worst
    strength: float
    #: The algorithm jitter, in (0, 1), larger values are worst
    jitter: float
    #: The algorithm complexity, in (0, 1), larger values are worst
    complexity: float

    def __init__(self,
                 name: str,
                 strength: float,
                 jitter: float,
                 complexity: float):
        """
        Create a mock algorithm description record.

        :param name: the algorithm name
        :param strength: the algorithm strength
        :param jitter: the algorithm jitter
        :param complexity: the algorithm complexity
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        if name != sanitize_name(name):
            raise ValueError(f"Invalid name '{name}'.")
        object.__setattr__(self, "name", name)

        if not isinstance(strength, float):
            raise type_error(strength, "strength", float)
        if (not isfinite(strength)) or (strength <= 0) or (strength >= 1):
            raise ValueError(
                f"strength must be in (0, 1), but is {strength}.")
        object.__setattr__(self, "strength", strength)

        if not isinstance(jitter, float):
            raise type_error(jitter, "jitter", float)
        if (not isfinite(jitter)) or (jitter <= 0) or (jitter >= 1):
            raise ValueError(
                f"jitter must be in (0, 1), but is {jitter}.")
        object.__setattr__(self, "jitter", jitter)

        if not isinstance(complexity, float):
            raise type_error(complexity, "complexity", float)
        if (not isfinite(complexity)) or (complexity <= 0) \
                or (complexity >= 1):
            raise ValueError(
                f"complexity must be in (0, 1), but is {complexity}.")
        object.__setattr__(self, "complexity", complexity)

    @staticmethod
    def create(n: int,
               forbidden: Optional[Any] = None,
               random: Generator = fixed_random_generator()) \
            -> Tuple['Algorithm', ...]:
        """
        Create a set of fixed mock algorithms.

        :param n: the number of algorithms to generate
        :param random: a random number generator
        :param forbidden: the forbidden names and strengths and so on
        :returns: a tuple of algorithms
        """
        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be > 0, but is {n}.")
        logger(f"now creating {n} algorithms.")

        not_allowed: Final[Set[Union[str, float]]] = \
            _make_not_allowed(forbidden)
        names: List[str] = []
        strengths: List[float] = []
        jitters: List[float] = []
        complexities: List[float] = []

        prefixes: Final[Tuple[str, ...]] = ('aco', 'bobyqa', 'cma-es', 'de',
                                            'ea', 'eda', 'ga', 'gp', 'hc',
                                            'ma', 'pso', 'rs', 'rw', 'sa',
                                            'umda')
        suffixes: Final[Tuple[str, ...]] = ("1swap", "2swap", "µ")

        max_name_len: int = int(max(2, ceil(n / 6)))
        trials: int = 0
        while len(names) < n:
            trials += 1
            if trials > 1000:
                trials = 1
                max_name_len += 1

            name_mode = random.integers(5)
            if name_mode < 2:
                nva = _random_name(int(3 + random.integers(max_name_len)),
                                   random)
                if nva in not_allowed:
                    continue
                if name_mode == 1:
                    nvb = suffixes[random.integers(len(suffixes))]
                    nv = f"{nva}_{nvb}"
                else:
                    nv = nva
                if nv in not_allowed:
                    continue
                not_allowed.add(nva)
                not_allowed.add(nv)
                names.append(nv)
                continue

            nva = prefixes[random.integers(len(prefixes))]
            if name_mode == 3:
                nvb = _random_name(int(3 + random.integers(
                    max_name_len)), random)
                if nvb in not_allowed:
                    continue
                nv = f"{nva}_{nvb}"
            elif name_mode == 4:
                nvb = suffixes[random.integers(len(suffixes))]
                nv = f"{nva}_{nvb}"
            else:
                nv = nva
                nvb = ""

            if nv in not_allowed:
                continue
            names.append(nv)
            not_allowed.add(nv)
            if name_mode == 3:
                not_allowed.add(nvb)

        limit: int = 2000
        trials = 0
        while len(strengths) < n:
            v: float = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5,
                                          scale=0.25) * limit) / limit
            if v not in not_allowed:
                strengths.append(v)
                not_allowed.add(v)

        limit = 2000
        trials = 0
        while len(jitters) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5,
                                          scale=0.2) * limit) / limit
            if v not in not_allowed:
                jitters.append(v)
                not_allowed.add(v)

        limit = 2000
        trials = 0
        while len(complexities) < n:
            v = -1
            while (v <= 0) or (v >= 1) or (not isfinite(v)):
                trials += 1
                if trials > 1000:
                    trials = 0
                    limit *= 2
                if random.integers(4) <= 0:
                    v = int(random.uniform(1, limit)) / limit
                else:
                    v = int(random.normal(loc=0.5,
                                          scale=0.2) * limit) / limit
            if v not in not_allowed:
                complexities.append(v)
                not_allowed.add(v)

        result: List[Algorithm] = []
        for i in range(n, 0, -1):
            result.append(Algorithm(
                name=names.pop(random.integers(i)),
                strength=strengths.pop(random.integers(i)),
                jitter=jitters.pop(random.integers(i)),
                complexity=complexities.pop(random.integers(i))))
        result.sort()

        logger(f"finished creating {n} algorithms.")
        return tuple(result)


@dataclass(frozen=True, init=False, order=True)
class BasePerformance:
    """An algorithm applied to a problem instance description record."""

    #: The algorithm.
    algorithm: Algorithm
    #: The problem instance
    instance: Instance
    #: The base performance, in (0, 1), larger values are worst
    performance: float
    #: The performance jitter, in (0, 1), larger values are worst
    jitter: float
    #: The time per FE, in (0, 1), larger values are worst
    speed: float

    def __init__(self,
                 algorithm: Algorithm,
                 instance: Instance,
                 performance: float,
                 jitter: float,
                 speed: float):
        """
        Create a mock algorithm-instance application description record.

        :param algorithm: the algorithm
        :param instance: the instance
        :param performance: the base performance
        :param jitter: the performance jitter
        :param speed: the time required per FE
        """
        if not isinstance(algorithm, Algorithm):
            raise type_error(algorithm, "algorithm", Algorithm)
        object.__setattr__(self, "algorithm", algorithm)
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        object.__setattr__(self, "instance", instance)

        if not isinstance(performance, float):
            raise type_error(performance, "performance", float)
        if (not isfinite(performance)) or (performance <= 0) \
                or (performance >= 1):
            raise ValueError(
                f"performance must be in (0, 1), but is {performance}.")
        object.__setattr__(self, "performance", performance)

        if not isinstance(jitter, float):
            raise type_error(jitter, "jitter", float)
        if (not isfinite(jitter)) or (jitter <= 0) or (jitter >= 1):
            raise ValueError(
                f"jitter must be in (0, 1), but is {jitter}.")
        object.__setattr__(self, "jitter", jitter)

        if not isinstance(speed, float):
            raise type_error(speed, "speed", float)
        if (not isfinite(speed)) or (speed <= 0) or (speed >= 1):
            raise ValueError(
                f"speed must be in (0, 1), but is {speed}.")
        object.__setattr__(self, "speed", speed)

    @staticmethod
    def create(instance: Instance,
               algorithm: Algorithm,
               random: Generator = fixed_random_generator()) \
            -> 'BasePerformance':
        """
        Compute the basic performance of an algorithm on a problem instance.

        :param instance: the instance tuple
        :param algorithm: the algorithm tuple
        :param random: the random number generator
        :returns: a tuple of the performance in (0, 1); bigger values are
            worse, and a jitter in (0, 1), where bigger values are worse
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        if not isinstance(algorithm, Algorithm):
            raise type_error(algorithm, "algorithm", Algorithm)
        logger("now creating base performance for algorithm "
               f"{algorithm.name} on instance {instance.name}.")

        perf: float = -1
        granularity: Final[int] = 2000
        while (perf <= 0) or (perf >= 1):
            if random.integers(20) <= 0:
                perf = random.uniform(low=0, high=1)
            else:
                perf = random.normal(
                    loc=0.5 * (instance.hardness + algorithm.strength),
                    scale=0.2 * (instance.jitter + algorithm.jitter))
            perf = int(perf * granularity) / granularity

        jit: float = -1
        while (jit <= 0) or (jit >= 1):
            if random.integers(15) <= 0:
                jit = random.uniform(low=0, high=1)
            else:
                jit = random.normal(
                    loc=0.5 * (instance.jitter + algorithm.jitter),
                    scale=0.2 * (instance.jitter + algorithm.jitter))
            jit = int(jit * granularity) / granularity

        speed: float = -1
        while (speed <= 0) or (speed >= 1):
            if random.integers(20) <= 0:
                speed = random.uniform(low=0, high=1)
            else:
                speed = random.normal(
                    loc=0.5 * (instance.scale + algorithm.complexity),
                    scale=0.2 * (instance.scale + algorithm.complexity))
            speed = int(speed * granularity) / granularity

        bp: Final[BasePerformance] = BasePerformance(algorithm=algorithm,
                                                     instance=instance,
                                                     performance=perf,
                                                     jitter=jit,
                                                     speed=speed)
        logger("finished base performance "
               f"{bp.algorithm.name}@{bp.instance.name}.")
        return bp


def get_run_seeds(instance: Instance, n_runs: int) -> Tuple[int, ...]:
    """
    Get the seeds for the runs.

    :param instance: the mock instance
    :param n_runs: the number of runs
    :returns: a tuple of seeds
    """
    if not isinstance(instance, Instance):
        raise type_error(instance, "instance", Instance)
    if not isinstance(n_runs, int):
        raise type_error(n_runs, "n_runs", int)
    if n_runs <= 0:
        raise ValueError(f"n_runs must be > 0, but is {n_runs}.")
    res: Final[Tuple[int, ...]] = tuple(sorted(rand_seeds_from_str(
        string=instance.name, n_seeds=n_runs)))
    logger(f"finished creating {n_runs} seeds for instance {instance.name}.")
    return res


@dataclass(frozen=True, init=False, order=True)
class Experiment:
    """An immutable experiment description."""

    #: The instances.
    instances: Tuple[Instance, ...]
    #: The algorithms.
    algorithms: Tuple[Algorithm, ...]
    #: The applications of the algorithms to the instances.
    applications: Tuple[BasePerformance, ...]
    #: The random seeds per instance.
    per_instance_seeds: Tuple[Tuple[int, ...]]
    #: the seeds per instance
    __seeds_per_inst: Dict[Union[str, Instance], Tuple[int, ...]]
    #: the performance per algorithm
    __perf_per_algo: Dict[Union[str, Algorithm], Tuple[BasePerformance, ...]]
    #: the performance per instance
    __perf_per_inst: Dict[Union[str, Instance], Tuple[BasePerformance, ...]]
    #: the algorithm by names
    __algo_by_name: Dict[str, Algorithm]
    #: the algorithm names
    algorithm_names: Tuple[str, ...]
    #: the instance by names
    __inst_by_name: Dict[str, Instance]
    #: the instance names
    instance_names: Tuple[str, ...]

    def __init__(self,
                 instances: Tuple[Instance, ...],
                 algorithms: Tuple[Algorithm, ...],
                 applications: Tuple[BasePerformance, ...],
                 per_instance_seeds: Tuple[Tuple[int, ...], ...]):
        """
        Create a mock experiment definition.

        :param algorithms: the algorithms
        :param instances: the instances
        :param applications: the applications of the algorithms to the
            instances
        :param per_instance_seeds: the seeds
        """
        if not isinstance(instances, tuple):
            raise type_error(instances, "instances", Tuple)
        if len(instances) <= 0:
            raise ValueError("instances must not be empty.")
        inst_bn: Dict[str, Instance] = {}
        for a in instances:
            if not isinstance(a, Instance):
                raise type_error(a, "element of instances", Instance)
            if a.name in inst_bn:
                raise ValueError(f"double instance name {a.name}.")
            inst_bn[a.name] = a
        object.__setattr__(self, "instances", instances)
        object.__setattr__(self, "_Experiment__inst_by_name", inst_bn)
        object.__setattr__(self, "instance_names",
                           tuple(sorted(inst_bn.keys())))

        if not isinstance(algorithms, tuple):
            raise type_error(algorithms, "algorithms", Tuple)
        if len(algorithms) <= 0:
            raise ValueError("algorithms must not be empty.")
        algo_bn: Dict[str, Algorithm] = {}
        for b in algorithms:
            if not isinstance(b, Algorithm):
                raise type_error(b, "element of algorithms", Algorithm)
            if b.name in algo_bn:
                raise ValueError(f"double algorithm name {b.name}.")
            if b.name in inst_bn:
                raise ValueError(f"instance/algorithm name {b.name} clash.")
            algo_bn[b.name] = b
        object.__setattr__(self, "algorithms", algorithms)
        object.__setattr__(self, "_Experiment__algo_by_name", algo_bn)
        object.__setattr__(self, "algorithm_names",
                           tuple(sorted(algo_bn.keys())))

        if not isinstance(applications, tuple):
            raise type_error(applications, "applications", Tuple)
        if len(applications) != len(algorithms) * len(instances):
            raise ValueError(
                f"There must be {len(algorithms) * len(instances)} "
                f"applications, but found {len(applications)}.")

        perf_per_inst: Dict[Instance, List[BasePerformance]] = {}
        perf_per_algo: Dict[Algorithm, List[BasePerformance]] = {}

        done: Set[str] = set()
        for c in applications:
            if not isinstance(c, BasePerformance):
                raise type_error(c, "element of applications", BasePerformance)
            s = c.algorithm.name + "+" + c.instance.name
            if s in done:
                raise ValueError(f"Encountered application {s} twice.")
            done.add(s)
            if c.algorithm in perf_per_algo:
                perf_per_algo[c.algorithm].append(c)
            else:
                perf_per_algo[c.algorithm] = [c]
            if c.instance in perf_per_inst:
                perf_per_inst[c.instance].append(c)
            else:
                perf_per_inst[c.instance] = [c]
        object.__setattr__(self, "applications", applications)

        pa: Dict[Union[str, Algorithm], Tuple[BasePerformance, ...]] = {}
        for ax in algorithms:
            lax: List[BasePerformance] = perf_per_algo[ax]
            lax.sort()
            pa[ax.name] = pa[ax] = tuple(lax)
        pi: Dict[Union[str, Instance], Tuple[BasePerformance, ...]] = {}
        for ix in instances:
            lix: List[BasePerformance] = perf_per_inst[ix]
            lix.sort()
            pi[ix.name] = pi[ix] = tuple(lix)
        object.__setattr__(self, "_Experiment__perf_per_algo", pa)
        object.__setattr__(self, "_Experiment__perf_per_inst", pi)

        if not isinstance(per_instance_seeds, tuple):
            raise type_error(per_instance_seeds, "per_instance_seeds", Tuple)
        if len(per_instance_seeds) != len(instances):
            raise ValueError(
                f"There must be one entry for each of the {len(instances)} "
                "instances, but per_instance_seeds only "
                f"has {len(per_instance_seeds)}.")
        xl: int = -1
        inst_seeds: Final[Dict[Union[str, Instance], Tuple[int, ...]]] = {}
        for idx, d in enumerate(per_instance_seeds):
            if not isinstance(d, tuple):
                raise type_error(d, "element of per_instance_seeds", Tuple)
            if len(d) <= 0:
                raise ValueError(f"there must be at least one per "
                                 f"instance seed, but found {len(d)}.")
            if xl < 0:
                xl = len(d)
            if len(d) != xl:
                raise ValueError(f"there must be {xl} per "
                                 f"instance seeds, but found {len(d)}.")
            for e in d:
                if not isinstance(e, int):
                    raise type_error(e, "element of seeds", int)
            inst_seeds[instances[idx]] = d
            inst_seeds[instances[idx].name] = d
        object.__setattr__(self, "per_instance_seeds", per_instance_seeds)
        object.__setattr__(self, "_Experiment__seeds_per_inst", inst_seeds)

    @staticmethod
    def create(n_instances: int,
               n_algorithms: int,
               n_runs: int,
               random: Generator = fixed_random_generator()) -> 'Experiment':
        """
        Create an experiment definition.

        :param n_instances: the number of instances
        :param n_algorithms: the number of algorithms
        :param n_runs: the number of per-instance runs
        :param random: the random number generator to use
        """
        if not isinstance(n_instances, int):
            raise type_error(n_instances, "n_instances", int)
        if n_instances <= 0:
            raise ValueError(
                f"n_instances must be > 0, but is {n_instances}.")
        if not isinstance(n_algorithms, int):
            raise type_error(n_algorithms, "n_algorithms", int)
        if n_algorithms <= 0:
            raise ValueError(
                f"n_algorithms must be > 0, but is {n_algorithms}.")
        if not isinstance(n_runs, int):
            raise type_error(n_runs, "n_runs", int)
        if n_algorithms <= 0:
            raise ValueError(
                f"n_runs must be > 0, but is {n_runs}.")
        logger(f"now creating mock experiment with {n_algorithms} algorithms "
               f"on {n_instances} instances for {n_runs} runs.")

        insts = Instance.create(n_instances, random=random)
        algos = Algorithm.create(n_algorithms, forbidden=insts, random=random)
        app = [BasePerformance.create(i, a, random)
               for i in insts for a in algos]
        app.sort()
        seeds = [get_run_seeds(i, n_runs) for i in insts]
        res: Final[Experiment] = Experiment(instances=insts,
                                            algorithms=algos,
                                            applications=tuple(app),
                                            per_instance_seeds=tuple(seeds))
        logger(f"finished creating mock experiment with {len(res.instances)} "
               f"instances, {len(res.algorithms)} algorithms, and "
               f"{len(res.per_instance_seeds[0])} runs per instance-"
               f"algorithm combination.")
        return res

    def seeds_for_instance(self, instance: Union[str, Instance]) \
            -> Tuple[int, ...]:
        """
        Get the seeds for the specified instance.

        :param instance: the instance
        :returns: the seeds
        """
        return self.__seeds_per_inst[instance]

    def instance_applications(self, instance: Union[str, Instance]) \
            -> Tuple[BasePerformance, ...]:
        """
        Get the applications of the algorithms to a specific instance.

        :param instance: the instance
        :returns: the applications
        """
        return self.__perf_per_inst[instance]

    def algorithm_applications(self, algorithm: Union[str, Algorithm]) \
            -> Tuple[BasePerformance, ...]:
        """
        Get the applications of an algorithm to the instances.

        :param algorithm: the algorithm
        :returns: the applications
        """
        return self.__perf_per_algo[algorithm]

    def get_algorithm(self, name: str) -> Algorithm:
        """
        Get an algorithm by name.

        :param name: the algorithm name
        :returns: the algorithm instance
        """
        return self.__algo_by_name[name]

    def get_instance(self, name: str) -> Instance:
        """
        Get an instance by name.

        :param name: the instance name
        :returns: the instance
        """
        return self.__inst_by_name[name]



"""Functions for testing multi-objective archive pruners."""
from typing import Final, List, Iterable, Optional, Callable

import numpy as np
from numpy.random import Generator, default_rng

from moptipy.api.mo_archive import MOArchivePruner, \
    check_mo_archive_pruner, MORecord
from moptipy.api.mo_problem import MOProblem
from moptipy.mock.mo_problem import MockMOProblem
from moptipy.tests.component import validate_component
from moptipy.utils.nputils import DEFAULT_NUMERICAL
from moptipy.utils.types import type_error


def validate_mo_archive_pruner(
        pruner_factory: Callable[[MOProblem], MOArchivePruner],
        dimensions: Iterable[int],
        dtypes: Iterable[np.dtype] = DEFAULT_NUMERICAL) -> None:
    """
    Check whether an object is a moptipy multi-objective optimization pruner.

    This method checks whether the class is correct and whether the pruning
    follows the general contract: Interesting records in the list to be pruned
    are moved to the front, the ones to discard are moved to the back. No
    record is lost and none is duplicated.

    :param pruner_factory: the creator for the multi-objective archive pruner
        to test
    :param dimensions: the dimensions to simulate
    :param dtypes: the dtypes to test on
    :raises ValueError: if `mo_problem` is not a valid
        :class:`~moptipy.api.mo_problem.MOProblem`
    :raises TypeError: if values of the wrong types are encountered
    """
    if not callable(pruner_factory):
        raise type_error(pruner_factory, "pruner_factory", call=True)
    if not isinstance(dimensions, Iterable):
        raise type_error(dimensions, "dimensions", Iterable)
    if not isinstance(dtypes, Iterable):
        raise type_error(dtypes, "dtypes", Iterable)

    tag: Final[str] = "bla"
    nothing: bool = True
    random: Final[Generator] = default_rng()
    for dim in dimensions:
        if not isinstance(dim, int):
            raise type_error(dim, "dimensions[i]", int)
        nothing = False
        for dt in dtypes:
            if not isinstance(dt, np.dtype):
                raise type_error(dt, "dt", np.dtype)

            mop: MockMOProblem = MockMOProblem.for_dtype(dim, dt)
            if not isinstance(mop, MockMOProblem):
                raise type_error(mop, "new mock problem", MockMOProblem)
            pruner = pruner_factory(mop)
            if not isinstance(pruner, MOArchivePruner):
                raise type_error(pruner, "pruner", MOArchivePruner)
            check_mo_archive_pruner(pruner)
            validate_component(pruner)

            alen = int(random.integers(2, 10))
            if not isinstance(alen, int):
                raise type_error(alen, "alen", int)
            amax = int(random.integers(1, alen + 1))
            if not isinstance(amax, int):
                raise type_error(amax, "amax", int)
            if not (0 < amax <= alen <= 10):
                raise ValueError(f"invalid amax={amax} and alen={alen}.")

            archive: List[MORecord] = []
            archivec: List[MORecord] = []

            for _ in range(alen):
                needed: bool = True
                rec: Optional[MORecord] = None
                fs: Optional[np.ndarray] = None
                while needed:   # we make sure that all records are unique
                    fs = np.empty(dim, dt)
                    if not isinstance(fs, np.ndarray):
                        raise type_error(fs, "fs", np.ndarray)
                    if len(fs) != dim:
                        raise ValueError(f"len(fs)={len(fs)}!=dim={dim}")
                    mop.sample(fs)
                    rec = MORecord(tag, fs)
                    if not isinstance(rec, MORecord):
                        raise type_error(rec, "rec", MORecord)
                    needed = False
                    for z in archive:
                        fs2 = z.fs
                        if np.array_equal(fs, fs2):
                            needed = True
                            break
                if (rec is None) or (fs is None):
                    raise ValueError("huh?")
                archive.append(rec)
                rec2 = MORecord(tag, fs.copy())
                if (rec.x != rec2.x) \
                        or (not np.array_equal(rec2.fs, rec.fs)):
                    raise ValueError(f"{rec} != {rec2}")
                archivec.append(rec2)

            # done creating archive and copy of archive

            if not isinstance(archive, List):
                raise type_error(archive, "archive", List)
            if not isinstance(archivec, List):
                raise type_error(archivec, "archivec", List)

            thelen: int = len(archive)
            if not isinstance(thelen, int):
                raise type_error(thelen, "len(archive)", int)
            if thelen != alen:
                raise ValueError(f"{alen} != len(archive)={len(archive)}?")

            thelen = len(archivec)
            if not isinstance(thelen, int):
                raise type_error(thelen, "len(archive)", int)
            if thelen != alen:
                raise ValueError(f"{alen} != len(archive)={len(archive)}?")

            # perform the pruning
            pruner.prune(archive, amax, len(archive))

            thelen = len(archive)
            if not isinstance(thelen, int):
                raise type_error(thelen, "len(archive)", int)
            if thelen != alen:
                raise ValueError(
                    f"pruning messed up archive len: {alen} != "
                    f"len(archive)={len(archive)}?")

            # make sure that no element was deleted or added
            for ii, a in enumerate(archive):
                if not isinstance(a, MORecord):
                    raise type_error(a, f"archive[{ii}]", MORecord)
                if a.x != tag:
                    raise ValueError(f"a.x={a.x}!='{tag}'")
                if not isinstance(a.fs, np.ndarray):
                    raise type_error(a.fs, "a.fs", np.ndarray)
                for idx, b in enumerate(archivec):
                    if np.array_equal(a.fs, b.fs):
                        if a.x != b.x:
                            raise ValueError(f"a.fs={a.fs}==b.fs, "
                                             f"but a.x={a.x}!=b.x={b.x}")
                        del archivec[idx]
                        break

    if nothing:
        raise ValueError("dimensions are empty!")
